#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2019 Erlend Ekern <dev@ekern.me>
#
# Distributed under terms of the MIT license.

import constants
import utils
from model import Model
from datetime import datetime
from types import SimpleNamespace
import time
import argparse
import os
import random
import re
import sys
import numpy as np
import tensorflow as tf
import matplotlib
# Set matplotlib backend to avoid script crashing
# ref. https://stackoverflow.com/questions/45993879/matplot-lib-fatal-io-error-25-inappropriate-ioctl-for-device-on-x-server-loc
matplotlib.use('Agg')


def set_seeds(seed):
    os.environ['PYTHONHASHSEED'] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    tf.set_random_seed(seed)


def validate_args(args):
    if args.mode in [constants.NON_SATURATING_GP, constants.WGAN_GP]:
        if args.discriminator_normalization == constants.BATCH_NORM:
            raise ValueError(
                f'{constants.BATCH_NORM} can not be used with mode {args.mode}'
            )
    if args.architecture_generator in [constants.CPPN, constants.CPPN_RESNET]:
        if args.generator_normalization == constants.BATCH_NORM:
            raise ValueError(
                f'{constants.BATCH_NORM} can not be used with generator architecture {args.architecture_generator}'
            )


def get_parsed_args():
    with open('./VERSION', 'r') as f:
        MODEL_VERSION = f.readline().strip()
    TIMESTAMP = datetime.now().strftime('%Y-%m-%d_%H:%M:%S')
    BASE_NAME = 'impgan'
    DEFAULT_NAME = f'{BASE_NAME}v{MODEL_VERSION}_{TIMESTAMP}'

    # TODO: Add all defaults to this namespace
    defaults = SimpleNamespace()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--name',
        type=str,
        help=f'The name of the model (default: {DEFAULT_NAME})',
        default=DEFAULT_NAME)
    parser.add_argument(
        '--load',
        type=str,
        help=
        'Directory with checkpoints or a specific checkpoint to restore from')
    parser.add_argument(
        '--mode',
        type=str,
        help=
        f'Which mode ({constants.NON_SATURATING}, {constants.WGAN}, {constants.WGAN_GP}) to use (default: {constants.WGAN_GP})',
        default=constants.WGAN_GP)
    parser.add_argument(
        '--dataset-name',
        type=str,
        default=constants.MNIST,
        help=
        f'The dataset ({constants.CELEB_A}, {constants.FASHION_MNIST}, {constants.MNIST}, {constants.CIFAR10}) to train on (default: {constants.MNIST})',
    )
    parser.add_argument(
        '--batch-size',
        type=int,
        help='The batch size (default: 256)',
        default=256)
    parser.add_argument(
        '--epochs',
        type=int,
        help='The number of epochs (default: 400)',
        default=400)
    parser.add_argument(
        '--save-interval',
        type=int,
        help='The model will be saved every nth epoch (default: 20)',
        default=20)
    parser.add_argument(
        '--sample-interval',
        type=int,
        help='The model will be sampled every nth epoch (default: 10)',
        default=10)
    parser.add_argument(
        '--seed',
        type=int,
        help='The random seed to use for NumPy and TensorFlow (default: 9)')
    parser.add_argument(
        '--sample',
        help=
        'Generate samples (all, codes, grid, stacked) only (default: None)',
        type=str,
        default=None)
    parser.add_argument(
        '--sample-num-stacked',
        help=
        'The number of stacked samples to generate when sampling outside of training (default: 1)',
        type=int,
        default=1)
    parser.add_argument(
        '--sample-dim',
        help=
        'The pixel dimension to use when generating high resolution samples (default: 1200)',
        type=int,
        default=1200)
    parser.add_argument(
        '--interpolate-dim',
        help=
        'The pixel dimension to use when interpolation between samples (default: 400)',
        type=int,
        default=400)
    parser.add_argument(
        '--interpolate-var',
        help=
        'The latent code variable (e.g., cat0, cat0_0-1, cat0_1-8, cont1, etc.) to interpolate over (default: cat0_0-1)',
        type=str,
        default='cat0_0-1')
    parser.add_argument(
        '--interpolate-frames',
        help=
        'The number of frames (read: interpolations) to create when interpolating between samples (default: 30)',
        type=int,
        default=30)
    parser.add_argument(
        '--sample-single-only',
        help=
        'Generate one high-res and one low-res sample only (no training) (default: False)',
        action='store_true',
        default=False)
    parser.add_argument(
        '--latent-size',
        type=int,
        help='The size of the latent vector Z (default: 128)',
        default=128)
    parser.add_argument(
        '--grid-size',
        type=int,
        help='The NxN-size of the sampling grid (default: 10)',
        default=10)
    parser.add_argument(
        '--gpu-index',
        type=int,
        help='The GPU to use if multiple are available (default: 0)',
        default=0)
    parser.add_argument(
        '--adam-beta1-auxiliary',
        type=float,
        help=
        'The beta1 parameter for the auxiliary net\'s Adam optimizer (default: 0.5)',
        default=0.5)
    parser.add_argument(
        '--adam-beta1-discriminator',
        type=float,
        help=
        'The beta1 parameter for the discriminator net\'s Adam optimizer (default: 0.5)',
        default=0.5)
    parser.add_argument(
        '--adam-beta1-generator',
        type=float,
        help=
        'The beta1 parameter for the generator\'s Adam optimizer (default: 0.5)',
        default=0.5)
    parser.add_argument(
        '--learning-rate-auxiliary',
        type=float,
        help='The learning rate for the auxiliary (Q) net (default: 0.0001)',
        default=0.0001)
    parser.add_argument(
        '--learning-rate-generator',
        type=float,
        help='The learning rate for the generator (default: 0.0001)',
        default=0.0001)
    parser.add_argument(
        '--learning-rate-discriminator',
        type=float,
        help='The learning rate for the discriminator (default: 0.0003)',
        default=0.0003)
    parser.add_argument(
        '--cppn-scale',
        type=float,
        help=
        'The scaling factor to use for the pixel interval in the CPPN (default: 10)'
    )
    parser.add_argument(
        '--smoothing-factor',
        type=float,
        help='Smoothing factor to decrease the discriminator\'s\
                confidence (default: 0.1)',
        default=0.1)
    parser.add_argument(
        '--verify-dataset',
        action='store_true',
        default=False,
        help='Verify dataset integrity on start (default: False)')
    parser.add_argument(
        '--interpolate',
        action='store_true',
        default=False,
        help='Interpolate between latent vectors and save the result\
                as a .GIF (default: False)')
    parser.add_argument(
        '--transform',
        action='store_true',
        default=False,
        help=
        'Gradually transform from low-resolution to high-resolution and save the result\
                as a .GIF (default: False)')
    parser.add_argument(
        '--transform-frames',
        default=30,
        type=int,
        help=
        'The number of frames to use when performing a transformation (default: 30)'
    )
    parser.add_argument(
        '--grid-interpolation',
        action='store_true',
        default=False,
        help='Interpolate between two grids of random latent vectors instead\
                of just two single vectors (default: False)')
    parser.add_argument(
        '--layer-sizes-auxiliary',
        type=int,
        nargs='+',
        help='The layers of the auxiliary net (default: 64)',
        default=[64])
    parser.add_argument(
        '--layer-sizes-generator',
        type=int,
        nargs='+',
        help='The layers of the generator (default: 128 128 128 128)',
        default=[128, 128, 128, 128])
    parser.add_argument(
        '--layer-sizes-discriminator',
        type=int,
        nargs='+',
        help=
        'The layers of the discriminator (default conv: 64 128 256, default mlp: 1024 512 256 128)',
        default=None)
    parser.add_argument(
        '--activation-functions-generator',
        type=str,
        nargs='+',
        help=
        f'The activation functions ({constants.ABS}, {constants.LEAKY_RELU}, {constants.RELU}, {constants.SIGN}, {constants.SIN}, {constants.TANH}) for the hidden layers of the generator (default: {constants.LEAKY_RELU} {constants.TANH} {constants.TANH} {constants.TANH})',
        default=[
            constants.LEAKY_RELU, constants.TANH, constants.TANH,
            constants.TANH
        ])
    parser.add_argument(
        '--categorical-sizes',
        type=int,
        nargs='+',
        help=
        'The sizes of the categorical codes in the auxiliary net (default: 10)',
        default=[10])
    parser.add_argument(
        '--num-continuous',
        type=int,
        help='The number of continuous codes in the auxiliary net (default: 2)',
        default=2)
    parser.add_argument(
        '--architecture-discriminator',
        type=str,
        help=
        f'The architecture ({constants.CONV}, {constants.MLP}) of the discriminator (default: {constants.CONV})',
        default=constants.CONV)
    parser.add_argument(
        '--architecture-generator',
        type=str,
        help=
        f'The architecture ({constants.CONV}, {constants.CPPN}, {constants.MLP}, {constants.CPPN_RESNET}) of the generator (default: {constants.CPPN})',
        default=constants.CPPN)
    parser.add_argument(
        '--kernel-size-discriminator',
        type=int,
        help=
        'The size of the kernel(s) used in the convolutional version of the discriminator (default: 4)',
        default=4)
    parser.add_argument(
        '--stride-discriminator',
        type=int,
        help=
        'The stride to use in the convolutional version of the discriminator (default: 2)',
        default=2)
    parser.add_argument(
        '--description',
        type=str,
        help='A short description of the current model (default: None)')
    parser.add_argument(
        '--cppn-stddev',
        type=float,
        help=
        'The standard deviation to use for the weight initialization of the hidden layers in the CPPN. Xavier init will be used if this is not set'
    )
    parser.add_argument(
        '--generator-resblock-intermediate-act-fn',
        type=str,
        default=constants.TANH,
        help=
        f'The activation function ({constants.ABS}, {constants.LEAKY_RELU}, {constants.RELU}, {constants.SIGN}, {constants.SIN}, {constants.TANH}) to use for the intermediate layers in the residual blocks of a Resnet CPPN (default: {constants.TANH})'
    )
    parser.add_argument(
        '--generator-resblock-out-act-fn',
        type=str,
        default=constants.LEAKY_RELU,
        help=
        f'The activation function ({constants.ABS}, {constants.LEAKY_RELU}, {constants.RELU}, {constants.SIGN}, {constants.SIN}, {constants.TANH}) to use for the output layer in the residual blocks of a Resnet CPPN (default: {constants.LEAKY_RELU})'
    )

    exclusive_resblock_group = parser.add_mutually_exclusive_group()
    exclusive_resblock_group.add_argument(
        '--generator-resblock-layer-size',
        type=int,
        default=5,
        help=
        'The size to use for all layers in the residual blocks of a ResNet CPPN (default: 5)'
    )
    exclusive_resblock_group.add_argument(
        '--generator-resblock-layer-sizes',
        type=int,
        nargs='+',
        help=
        'The size of each layer in the residual blocks of a ResNet CPPN. You need to explicitly set the size of every layer if using this option. See --generator-resblock-layer-size for an alternative that uses the same size for all layers. (default: [5, 5, 5, 5])'
    )
    parser.add_argument(
        '--generator-resblock-num-layers',
        type=int,
        default=4,
        help=
        'The number of layers in each residual block of a Resnet CPPN (default: 4)'
    )
    parser.add_argument(
        '--generator-num-resblocks',
        type=int,
        default=20,
        help=
        'The number of residual blocks to use in a Resnet CPPN (default: 20)')
    parser.add_argument(
        '--cppn-exclude-radius',
        action='store_true',
        default=False,
        help=
        'Whether the radius term should be removed as an input to the CPPN (default: False)'
    )
    parser.add_argument(
        '--generator-normalization',
        type=lambda x: None if x == 'None' else x,
        default=None,
        help=
        f'The normalization scheme ({constants.BATCH_NORM}, None) to use for the generator (default: None)'
    )
    parser.add_argument(
        '--discriminator-normalization',
        type=lambda x: None if x == 'None' else x,
        default=constants.SPECTRAL_NORM,
        help=
        f'The normalization scheme ({constants.LAYER_NORM}, {constants.SPECTRAL_NORM}, None) to use for the discriminator (default: {constants.SPECTRAL_NORM})'
    )
    parser.add_argument(
        '--auxiliary-normalization',
        type=lambda x: None if x == 'None' else x,
        default=constants.SPECTRAL_NORM,
        help=
        f'The normalization scheme ({constants.LAYER_NORM}, {constants.SPECTRAL_NORM}, None) to use for the auxiliary net (default: {constants.SPECTRAL_NORM})'
    )
    args = parser.parse_args()

    validate_args(args)

    # Set default layer sizes for the discriminator depending on the architecture
    if args.layer_sizes_discriminator is None:
        if args.architecture_discriminator == constants.CONV:
            args.layer_sizes_discriminator = [64, 128, 256]
        elif args.architecture_discriminator == constants.MLP:
            args.layer_sizes_discriminator = [1024, 512, 256, 128]

    if args.architecture_generator == constants.CPPN_RESNET:
        if args.generator_resblock_layer_sizes is None:
            args.generator_resblock_layer_sizes = [
                args.generator_resblock_layer_size
            ] * args.generator_resblock_num_layers
        else:
            args.generator_resblock_num_layers = len(
                args.generator_resblock_layer_sizes)

    if args.load:
        matches = re.match(r'^.*save/([^/]+)/.*$', args.load)
        groups = matches.groups() if matches else []
        if len(groups) == 1:
            # We found a valid model
            # Read its configuration
            args.name = groups[0]
            config = utils.read_yaml(filename=f'save/{args.name}/model.yml')
            set_args_from_config(args, config)
        else:
            sys.exit('Could not extract name of restored model')

    # Set default seed if necessary
    if args.seed is None:
        args.seed = 9

    # Set default scale if necessary
    if args.cppn_scale is None:
        args.cppn_scale = 10

    return args


def set_args_from_config(args, config):
    # TODO: Avoid directly mutating args here
    args.adam_beta1_auxiliary = config['auxiliary']['adam_beta1']
    args.adam_beta1_discriminator = config['discriminator']['adam_beta1']
    args.adam_beta1_generator = config['generator']['adam_beta1']
    args.layer_sizes_auxiliary = config['auxiliary']['layer_sizes']
    args.layer_sizes_discriminator = config['discriminator']['layer_sizes']
    args.learning_rate_auxiliary = config['auxiliary']['learning_rate']
    args.learning_rate_discriminator = config['discriminator']['learning_rate']
    args.learning_rate_generator = config['generator']['learning_rate']
    args.categorical_sizes = config['auxiliary']['categorical_sizes']
    args.auxiliary_normalization = config['auxiliary']['normalization']
    args.discriminator_normalization = config['discriminator']['normalization']
    args.generator_normalization = config['generator']['normalization']
    args.num_continuous = config['auxiliary']['num_continuous']
    args.dataset_name = config['dataset_name']
    args.latent_size = config['generator']['latent_size']
    args.architecture_discriminator = config['discriminator']['architecture']
    args.architecture_generator = config['generator']['architecture']
    if args.architecture_discriminator == constants.CONV:
        args.kernel_size_discriminator = config['discriminator']['kernel_size']
        args.stride_discriminator = config['discriminator']['stride']
    if args.architecture_generator == constants.CPPN_RESNET:
        args.generator_num_resblocks = config['generator']['num_resblocks']
        args.generator_resblock_intermediate_act_fn = config['generator'][
            'resblock']['intermediate_act_fn']
        args.generator_resblock_out_act_fn = config['generator']['resblock'][
            'out_act_fn']
        args.generator_resblock_num_layers = config['generator']['resblock'][
            'num_layers']
        args.generator_resblock_layer_sizes = config['generator']['resblock'][
            'layer_sizes']
    elif args.architecture_generator == constants.CPPN or args.architecture_generator == constants.MLP:
        args.activation_functions_generator = config['generator'][
            'activation_functions']
        args.layer_sizes_generator = config['generator']['layer_sizes']

    if args.architecture_generator in [constants.CPPN, constants.CPPN_RESNET]:
        args.cppn_exclude_radius = config['generator']['exclude_radius']
        args.cppn_stddev = config['generator']['stddev']

        # Use scale from command-line argument if it is supplied when loading a model
        if args.cppn_scale is None:
            args.cppn_scale = config['generator']['scale']

    args.mode = config['mode']

    # Use seed from command-line argument if it is supplied when loading a model
    if args.seed is None:
        args.seed = config['seed']
    args.batch_size = config['batch_size']


def should_train(args):
    if args.load:
        return not (args.sample or args.interpolate or args.transform)
    return True


def main():
    args = get_parsed_args()
    train = should_train(args)

    # Create missing directories if necessary
    if args.load is None:
        save_dir = f'save/{args.name}'
        if not os.path.exists(save_dir):
            os.makedirs(f'{save_dir}/inference')
            os.makedirs(f'{save_dir}/checkpoints')
            os.makedirs(f'{save_dir}/summary')
        else:
            sys.exit(
                f'A model with the name \'{args.name}\' already exists! Exiting ...'
            )

    model_dir = f'save/{args.name}'
    description_file = f'{model_dir}/description.txt'
    log_file = f'save/{args.name}/log.txt'
    checkpoints_dir = f'save/{args.name}/checkpoints'
    inference_dir = f'save/{args.name}/inference'

    # Set seeds after we potentially have read config from file
    set_seeds(args.seed)

    model = Model(
        activation_functions_generator=args.activation_functions_generator,
        adam_beta1_auxiliary=args.adam_beta1_auxiliary,
        adam_beta1_discriminator=args.adam_beta1_discriminator,
        adam_beta1_generator=args.adam_beta1_generator,
        architecture_discriminator=args.architecture_discriminator,
        architecture_generator=args.architecture_generator,
        auxiliary_normalization=args.auxiliary_normalization,
        batch_size=args.batch_size,
        categorical_sizes=args.categorical_sizes,
        dataset_name=args.dataset_name,
        discriminator_normalization=args.discriminator_normalization,
        generator_normalization=args.generator_normalization,
        generator_num_resblocks=args.generator_num_resblocks,
        generator_resblock_intermediate_act_fn=args.
        generator_resblock_intermediate_act_fn,
        generator_resblock_out_act_fn=args.generator_resblock_out_act_fn,
        generator_resblock_layer_sizes=args.generator_resblock_layer_sizes,
        generator_resblock_num_layers=args.generator_resblock_num_layers,
        gpu_index=args.gpu_index,
        kernel_size_discriminator=args.kernel_size_discriminator,
        latent_size=args.latent_size,
        layer_sizes_auxiliary=args.layer_sizes_auxiliary,
        layer_sizes_discriminator=args.layer_sizes_discriminator,
        layer_sizes_generator=args.layer_sizes_generator,
        learning_rate_auxiliary=args.learning_rate_auxiliary,
        learning_rate_discriminator=args.learning_rate_discriminator,
        learning_rate_generator=args.learning_rate_generator,
        mode=args.mode,
        model_dir=model_dir,
        name=args.name,
        num_continuous=args.num_continuous,
        sampling_grid_size=args.grid_size,
        seed=args.seed,
        cppn_scale=args.cppn_scale,
        cppn_exclude_radius=args.cppn_exclude_radius,
        cppn_stddev=args.cppn_stddev,
        stride_discriminator=args.stride_discriminator,
        train=train)

    if args.load:
        model.restore(args.load)
    if not train:
        if args.sample == 'all':
            if args.architecture_generator in [
                    constants.CPPN, constants.CPPN_RESNET
            ]:
                model.sample_stacked(
                    img_dim_high=args.sample_dim,
                    num_samples=args.sample_num_stacked)
            model.sample_codes()
            model.sample_grid()
        elif args.sample == 'codes':
            model.sample_codes()
        elif args.sample == 'grid':
            model.sample_grid()
        elif args.sample == 'stacked':
            model.sample_stacked(
                img_dim_high=args.sample_dim,
                num_samples=args.sample_num_stacked)

        if args.interpolate:
            if args.architecture_generator in [
                    constants.CPPN, constants.CPPN_RESNET
            ]:
                model.interpolate(
                    img_dim=args.interpolate_dim,
                    num_interpolations=args.interpolate_frames,
                    interpolate_var=args.interpolate_var)
            else:
                model.interpolate(
                    num_interpolations=args.interpolate_frames,
                    interpolate_var=args.interpolate_var)
        if args.transform:
            if args.architecture_generator not in [
                    constants.CPPN, constants.CPPN_RESNET
            ]:
                raise ValueError(
                    'The generator must be a CPPN in order to do a transformation'
                )
            model.transform(num_frames=args.transform_frames)

        model.close()
        sys.exit()

    # Write model description to file
    if args.description:
        if os.path.exists(description_file):
            raise ValueError(
                f'Description file {description_file} already exists')
        print(args.description, file=open(description_file, 'w'))

    for epoch in range(args.epochs):
        start = time.time()
        loss_discriminator, loss_generator, loss_auxiliary, step = model.train_epoch(
        )
        end = time.time()
        model.log(
            f'Epoch: {epoch}/{args.epochs} | Step: {step} | D_loss: {loss_discriminator:.4f} | G_loss: {loss_generator:.4f} | Q_loss: {loss_auxiliary:.4f} | {(end - start):.1f}s'
        )
        if epoch % args.sample_interval == 0:
            start = time.time()
            model.sample_grid()
            if args.architecture_generator == constants.CPPN or args.architecture_generator == constants.CPPN_RESNET:
                model.sample_stacked()
            model.sample_codes()
            end = time.time()
            model.log(f'Sampling took: {(end - start):.1f}s')
        if epoch % args.save_interval == 0:
            start = time.time()
            model.save()
            end = time.time()
            model.log(
                f"Saving model: \'{args.name}\' took {(end - start):.1f}s")
    model.save()
    model.close()


if __name__ == '__main__':
    main()
